/*

  P-Code interpreter (to run the apple pascal system)
  Copyright (C) 2000 Mario Klebsch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <stdarg.h>
#include <setjmp.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "psystem.h"
#include "Memory.h"
#include "Stack.h"
#include "Sets.h"
#include "Array.h"
#include "Diskio.h"

#define IXP_COMPATIBILITY

word DisasmP	(char *Buffer,
		 word IpcBase,
		 word Ipc,
		 word JTab,
		 word Sp);

static FILE *Out=NULL;

#define APPLE_HEAP_BOT		0x0804
#define APPLE_KP_TOP		0xfe7c
#define APPLE_SEG0_LOAD_GAP	0x450a
#define APPLE_SYSCOM		0xbdde

/* #define KP_TOP			0xfe80 */
#define KP_TOP			0xe000
#define HEAP_BOT		0x1000

#define NUMBER(a)	(sizeof(a)/sizeof(a[0]))

typedef struct SegDict
{
  int		UseCount;
  word		OldKp;
  word		Seg;
  word		SegBase;
} SegDict_t;

SegDict_t	SegDict[32];

#define MS_KP		-1
#define MS_STAT		0
#define MS_DYN		1
#define MS_JTAB		2
#define MS_SEG		3
#define MS_IPC		4
#define	MS_SP		5
#define MS_VAR		5		/* Var-Offset counts from 1.. */
#define MS_FRAME_SIZE	6

/* Official P-Machine registers */
word	Sp;
word	Ipc;
word	IpcBase;
word	Seg;
word	JTab;
word	Kp;
word	Mp;
word	Np;
word	Base;

word	Syscom;

/* Flags */

#ifndef WORD_MEMORY
int	AppleCompatibility=0;
#endif

/* Additional Bookkeeping */
static word CurrentIpc;
static word InitialMp;
unsigned int	Level=0;
unsigned int	TraceLevel=0x7fff;
jmp_buf		ProcessNextInstrunction;

void warning(char *Msg, ...)
{
  va_list ap;
  char	Buffer[512];
  va_start(ap, Msg);
  vsnprintf(Buffer, sizeof(Buffer), Msg, ap);
  va_end(ap);
  fprintf(stderr,"warning: %s\n", Buffer);
  /*  TraceLevel=0x7fff;*/
}

void DumpCore(void)
{
  FILE		*f;
  if ((f=fopen("ucsd.core","w")))
    {
      MemDump(f,0,0xffff);
      fclose(f);
    }
  else
    warning("DumpCore: unable to create core dump");
}

void panic(char *Msg, ...)
{
  va_list	ap;
  char		Buffer[512];

  va_start(ap, Msg);
  vsnprintf(Buffer, sizeof(Buffer), Msg, ap);
  va_end(ap);
  fprintf(stderr,"panic: %s\n", Buffer);
  DumpCore();
  abort();
}

word FetchB()
{
  byte b;
  b = MemRdByte(IpcBase, Ipc++);
  if (b&0x80)
    return ( (word)((b&0x7f)<<8) + (word)MemRdByte(IpcBase, Ipc++) );
  else
    return ( (word)b );
}

inline word FetchW()
{
  word	w;
  w =  MemRdByte(IpcBase, Ipc++);
  w += (MemRdByte(IpcBase, Ipc++)<<8);
  return(w);
}

inline word FetchUB()
{
  return ( (word)MemRdByte(IpcBase, Ipc++) );
}

int AppleHack1(void)
{
  word Save0=Ipc;

  if (FetchUB() == 145)				/* NGI */
    {
      word OpCode=FetchUB();
      if (OpCode == 171)			/* SRO */
	{
	  word Var=FetchB();			/* Parameter SRO */
	  OpCode=FetchUB();
	  if ((((OpCode   == 169) &&		/* LDO n */
		(FetchB() == Var)) ||
	       ((Var>=1) && (Var<=16) &&
		(OpCode   == 231+Var))) &&	/* SLDO n */
	      (FetchUB()  ==   0) &&		/* SLCD	0 */
	      (FetchUB()  == 190) )		/* LDB */
	    return(1);
	}
      else if (OpCode == 204)			/* STL */
	{
	  word Var=FetchB();			/* Parameter STL */
	  OpCode=FetchUB();
	  if ((((OpCode   == 202) &&		/* LDL  n */
		(FetchB() == Var)) ||
	       ((Var>=1) && (Var<=16) &&
		(OpCode   == 215+Var))) &&	/* SLDL n */
	      (FetchUB()  ==   0) &&		/* SLCD	0 */
	      (FetchUB()  == 190) )		/* LDB */
	    return(1);
	}
    }
  Ipc=Save0;
  return(0);
}

int AppleHack2(void)
{
  word Save=Ipc;
  word Var;

  if ( (FetchUB() == 145) &&		/* NGI */
       (FetchUB() == 171) &&		/* SRO */
       (Var=FetchB()) &&		/* Parameter SRO */
       (FetchUB() == 169) &&		/* LDO n */
       (FetchB()  == Var) &&
       (FetchUB() == 6) &&		/* SLDC 6 */
       (FetchUB() == 192) &&		/* IXP 16,1 */
       (FetchUB() == 16) &&
       (FetchUB() == 1) &&
       (FetchUB() == 186))		/* LDP */
    return(1);
  Ipc=Save;
  return(0);
}

static inline word SelfRelPtr(word Addr)
{
#ifdef WORD_MEMORY
  return(Addr-MemRd(Addr)/2);
#else
  return(Addr-MemRd(Addr));
#endif
}

inline byte SegNumProc(word Seg)
{
  return(MemRd(Seg)>>8);
}

inline byte SegNumber(word Seg)
{
  return(MemRd(Seg)&0xff);
}

word Proc(word Seg, byte ProcNr)
{
  PointerCheck(Seg);
  if ((ProcNr<1) || (ProcNr>SegNumProc(Seg)) )
    panic("Proc: Illegal Procedure Number %d",ProcNr);
  return(SelfRelPtr(WordIndexed(Seg,-ProcNr)));
}

inline signed char ProcNumber(word JTab)
{
  PointerCheck(JTab);
  return (MemRd(JTab)&0xff);
}

inline signed char ProcLexLevel(word JTab)
{
  PointerCheck(JTab);
  return (MemRd(JTab)>>8);
}

static inline word ProcBase(word JTab)
{
  PointerCheck(JTab);
  return (SelfRelPtr(WordIndexed(JTab,-1)));
}

inline word ProcExitIpc(word JTab)
{
  PointerCheck(JTab);
  return (MemRd(WordIndexed(JTab,-1))-
	  MemRd(WordIndexed(JTab,-2))-2);
}

inline word ProcParamSize(word JTab)
{
  PointerCheck(JTab);
  return (MemRd(WordIndexed(JTab,-3)));
}

inline word ProcDataSize(word JTab)
{
  PointerCheck(JTab);
  return (MemRd(WordIndexed(JTab,-4)));
}

inline word ProcJTab(word JTab)
{
  PointerCheck(JTab);
  return(JTab-10);
}

inline word Boolean(word i)
{
  return(i?1:0);
}

char *PStr(word w)
{
  static char	Buffer[256];
  int		len=MemRdByte(w,0);
  char		*p=Buffer;
  int		i;

  for (i=0; i<len; i++)
    *p++=MemRdByte(w, i+1);
  *p++='\0';
  return(Buffer);
}

char *MultipleWords(word Addr, word Len)
{
  static char	Buffer[5*256+1];
  char		*p=Buffer;
  while (Len--)
    {
      sprintf(p,",%04x", MemRd(Addr));
      p+=strlen(p);
      Addr=WordIndexed(Addr, 1);
    }
  return(Buffer);
}

inline word LocalAddr(word Offset)
{
  return(WordIndexed(Mp, MS_VAR+Offset));
}

inline word GlobalAddr(word Offset)
{
  return(WordIndexed(Base,MS_VAR+Offset));
}

inline word Intermediate(byte Count)
{
  word p;
  for (p=Mp;Count;Count--)
    p=MemRd(WordIndexed(p,MS_STAT));
  return(p);
}

inline word IntermediateAddr(word Offset, byte Count)
{
  return( WordIndexed(Intermediate(Count), MS_VAR+Offset) );
}

inline word ExtendedAddr(word Offset, byte SegNo)
{
  assert(SegNo<NUMBER(SegDict));
  return(WordIndexed(SegDict[SegNo].Seg, Offset));
}

void MoveLeft(word Dst, Integer DstOffset,
	      word Src, Integer SrcOffset, word Len)
{
  while (Len--)
    MemWrByte(Dst, DstOffset++,MemRdByte(Src, SrcOffset++));
}

void MoveRight(word Dst, Integer DstOffset,
	       word Src, Integer SrcOffset, word Len)
{
  SrcOffset+=Len;
  DstOffset+=Len;
  while (Len--)
    MemWrByte(Dst, --DstOffset, MemRdByte(Src,--SrcOffset));
}

word jump(signed char disp)
{
  if (disp >=0)
    return(Ipc+disp);
  disp=-disp;
#ifdef WORD_MEMORY
  return(MemRd(WordIndexed(JTab, -1))+2-
	 (MemRd(JTab-disp/2)+disp));
#else
  return(MemRd(WordIndexed(JTab, -1))+2-
	 (MemRd(JTab-disp)+disp));
#endif
}

word StaticLink(word NewSeg, byte ProcNo)
{
  word NewJTab=Proc(NewSeg, ProcNo);

  if (!ProcNumber(NewJTab))
    return(NIL);
  return(Intermediate(ProcLexLevel(JTab)-
		      ProcLexLevel(NewJTab)+1));
}

void CspLoadSegment(byte SegNo)
{
  assert (SegNo<NUMBER(SegDict));
  if (!SegDict[SegNo].UseCount)
    {
      word SegUnit  = MemRd(SEG_UNIT(SegNo));
      word SegBlock = MemRd(SEG_BLOCK(SegNo));
      word SegSize  = MemRd(SEG_SIZE(SegNo));

      assert (!(SegSize&1));

      SegDict[SegNo].OldKp=Kp;
#ifdef WORD_MEMORY
      Kp-=SegSize/2;
#else
      Kp-=SegSize;
#endif
      SegDict[SegNo].SegBase=Kp;
      if (SegBlock)
	{
	  SegDict[SegNo].Seg=WordIndexed(SegDict[SegNo].OldKp,-1);
	  UnitRead(SegUnit, Kp, 0, SegSize, SegBlock, 0);
	  if (MemRd(IORSLT))
	    XeqError(9);
	}
      else			/* Data-Segment			*/
	{
	  SegDict[SegNo].Seg=WordIndexed(Kp,-1);
	}
    }
  SegDict[SegNo].UseCount++;
}

void CspUnloadSegment(byte SegNo)
{
  SegDict[SegNo].UseCount--;
  if (!SegDict[SegNo].UseCount)
    {
      Kp=SegDict[SegNo].OldKp;
      SegDict[SegNo].OldKp = 0;
      SegDict[SegNo].Seg   = 0;
    }
}

void ClrGDirP(void)
{
  word GDirP=MemRd(GDIRP);
  if (GDirP!=NIL)
    {
      Np=GDirP;
      MemWr(GDIRP, NIL);
    }
}

void StackCheck(void)
{
  if (Np>=Kp)
    {
      MemWr(GDIRP, NIL);
      Kp=0x8000;
      Np=0x6200;
      XeqError(4);
    }
}

int call(word NewSeg, byte ProcNr, word StaticLink)
{
  word	NewJTab	  = Proc(NewSeg, ProcNr);
  word	DataSize  = ProcDataSize(NewJTab);
  word	ParamSize = ProcParamSize(NewJTab);
  word	NewMp	  = WordIndexed(Kp, -(DataSize + ParamSize)/2);

  if (!ProcNumber(NewJTab))
    {
      ProcessNative(NewJTab);
      return(1);
    }

  assert(!(ParamSize&1));

  MoveLeft(NewMp, 0, Sp, 0, ParamSize);
  Sp=WordIndexed(Sp,ParamSize/2);

  NewMp = WordIndexed(NewMp, -MS_FRAME_SIZE);
  MemWr(WordIndexed(NewMp,MS_KP),   Kp);
  MemWr(WordIndexed(NewMp,MS_STAT), StaticLink);
  MemWr(WordIndexed(NewMp,MS_DYN),  Mp);
  MemWr(WordIndexed(NewMp,MS_JTAB), JTab);
  MemWr(WordIndexed(NewMp,MS_SEG),  Seg);
  MemWr(WordIndexed(NewMp,MS_IPC),  Ipc);
  MemWr(WordIndexed(NewMp,MS_SP),   Sp);

  Kp      = WordIndexed(NewMp, -1);	/* Kleiner Hack :-( */
  Mp      = NewMp;
  MemWr(LASTMP, Mp);
  Seg     = NewSeg;
  JTab    = NewJTab;
  IpcBase = ProcBase(JTab);
  Ipc     = 0;
  if (ProcLexLevel(JTab) <= 0)
    {
      Push(Base);
      Base=Mp;
      MemWr(STKBASE,Base);
    }
  Level++;
  StackCheck();
  return(0);
}

void ret(byte n)
{
  word	OldMp=Mp;
  byte	OldSegNo=SegNumber(Seg);

  Sp  = MemRd(WordIndexed(OldMp,MS_SP));
  while (n>0)
    Push(MemRd(LocalAddr(n--)));

  Kp      = MemRd(WordIndexed(OldMp,MS_KP));
  Mp      = MemRd(WordIndexed(OldMp,MS_DYN));
  MemWr(LASTMP, Mp);
  JTab    = MemRd(WordIndexed(OldMp,MS_JTAB));
  IpcBase = ProcBase(JTab);
  Seg     = MemRd(WordIndexed(OldMp,MS_SEG));
  Ipc     = MemRd(WordIndexed(OldMp,MS_IPC));

  if (OldSegNo != SegNumber(Seg))
    if (OldSegNo)			/* Segment 0 wird nicht verwaltet   */
      CspUnloadSegment(OldSegNo);
  Level--;
  StackCheck();
}

void XeqError(word err)
{
  static int Flag=0;
  word	NewSeg=SegDict[0].Seg;

  if (Flag)
    panic("XeqError: recursion");
  Flag++;

  MemWr(XEQERR,  err);
  MemWr(BOMIPC,  CurrentIpc);
  MemWr(BOMPROC, ProcNumber(JTab));
  MemWr(BOMSEG,  SegNumber(Seg));

  call(NewSeg, 2, InitialMp );
#ifdef XXX
  TraceLevel=0x7fff;
  warning ("XeqError(%d)", err);
#endif
  Flag--;
  longjmp(ProcessNextInstrunction, 0);
}

#define TRACE(a) Trace a

void ShowMem(word Start, word End)
{
  for ( ;Start<End; Start=WordIndexed(Start,1))
    fprintf(stderr," %d(%x)",MemRd(Start), MemRd(Start));
  fprintf(stderr,"\n");
}

void List(word JTab)
{
  word	IpcBase=ProcBase(JTab);
  word	Ipc=0;
  char	Buffer[1024];
  printf("Params: %d, Vars: %d\n",
	 ProcParamSize(JTab)/2,
	 ProcDataSize(JTab)/2);
  while (WordIndexed(IpcBase, Ipc/2)<JTab)
    {
      word OpCode=MemRdByte(IpcBase, Ipc);
      sprintf(Buffer,"%d:	", Ipc);
      Ipc=DisasmP(Buffer+strlen(Buffer), IpcBase, Ipc, JTab, 0);
      printf("%s\n", Buffer);
      if ( (OpCode==173) ||
	   (OpCode==193) ||
	   (OpCode==214) )
	return;
    }
}

void Debugging(char *s,...)
{
  va_list	arg;
  char		prompt[64];
  char		Buffer[256];
  int		from,to;
  char		Buf[10];
  va_start (arg, s);			/* Format-String auswerten	    */
  vsnprintf (prompt, sizeof(prompt), s, arg);
  va_end(arg);

  do
    {
      Buffer[0]='\0';
      fprintf(stderr,"%s", prompt);
      gets(Buffer);
      switch (Buffer[0])
	{
	case 'p':		/* print stack */
	  fprintf(stderr,"Stack:");
	  ShowMem(Sp,SP_TOP);
	  break;
	case 'd':
	  switch (sscanf(Buffer, "%10s %x %x", Buf, &from, &to))
	    {
	      /*
	    case 1:
	      from=NextDumpAddr;
	      */
	    case 2:
	      to=from+0x80;
	    case 3:
	      MemDump(stdout, from,to);
	      break;
	    default:
	      fprintf(stderr,"d <from> [<to>]\n");
	    }
	  break;
	case 'l':
	  {
	    int SegNo;
	    int ProcNo;
	    switch (sscanf(Buffer, "%10s %d %d", Buf, &SegNo, &ProcNo))
	      {
	      case 2:
		ProcNo=SegNo;
		SegNo=SegNumber(Seg);
	      case 3:
		if (SegNo<NUMBER(SegDict))
		  {
		    CspLoadSegment(SegNo);
		    List(Proc(SegDict[SegNo].Seg, ProcNo));
		    CspUnloadSegment(SegNo);
		  }
		break;
	      default:
		fprintf(stderr,"l [<SegNo>] <ProcNo>\n");
	      }
	  }
	  break;
	case 't':
	  {
	    word s=Seg;
	    word j=JTab;
	    word m=Mp;
	    word i=Ipc;
	    
	    while (1)
	      {
		word w;
		fprintf(stderr,"\ns%d, p%d, %4d:\n",
			SegNumber(s), ProcNumber(j), i);
		w=WordIndexed(m, MS_VAR);
		MemDump(stderr, w,w+ProcParamSize(j)+ProcDataSize(j));

		if (ProcLexLevel(j)<0)
		  break;
		j = MemRd(WordIndexed(m,MS_JTAB));
		s = MemRd(WordIndexed(m,MS_SEG));
		i = MemRd(WordIndexed(m,MS_IPC));
		m = MemRd(WordIndexed(m,MS_DYN));
	      }
	  }
	  MemDump(stdout, Kp, 0xb000);
	  break;
	case 'v':
	  MemDump(stdout, WordIndexed(Mp,MS_VAR), WordIndexed(Mp,MS_VAR)+ProcDataSize(JTab)+ProcParamSize(JTab));
	  break;
	case 'g':
	  TraceLevel=0;
	  return;
	case 'n':
	  TraceLevel=Level;
	  return;
	case 'f':
	  TraceLevel=Level-1;
	  return;
	case 'r':
	  fprintf(stderr,"Sp=%04x, Kp=%04x, Mp=%04x, Base=%04x, Seg=%04x, JTab=%04x, Np=%04x\n",
		  Sp, Kp, Mp, Base, Seg, JTab, Np);
	  break;
	case 'q':
	  if (Out)
	    fclose(Out);
	  exit(0);
	  break;
	}
    } while (Buffer[0]);
}

void Trace(char *s,...)
{
  va_list arg;
  char    msg[64];
  
  if (Level>TraceLevel)
    return;
  TraceLevel=0x7fff;

  va_start (arg, s);			/* Format-String auswerten	    */
  vsnprintf (msg, sizeof(msg), s, arg);
  va_end(arg);
  Debugging("s%d, p%d, %4d:	%s	> ", SegNumber(Seg), ProcNumber(JTab), 
	    CurrentIpc, msg);
}

word Translate(word Value)
{
#ifdef WORD_MEMORY
  if (Value>KP_TOP)
    ;
  else if (Value>0x8000)
    Value = (Value-KP_TOP)*2+KP_TOP;
  else if (Value>0x7f00)
    ;
  else if (Value >HEAP_BOT)
    Value = (Value-HEAP_BOT)*2+HEAP_BOT;
#endif
  return(Value);
}

void Tracer(void)
{
  char	Buffer[64000];
  char	StackBuf[1024];
  char	*p=StackBuf;
  word  w=Sp;

  *p='\0';
  while (w<SP_TOP)
    {
      word Value=MemRd(w);

      sprintf(p,"%04x ",Translate(Value));
#ifdef XXX
      if (Value>=0x200)
	sprintf(p,".... ");
      else
	sprintf(p,"%04x ",Value);
#endif
      p+=strlen(p);
      w=WordIndexed(w,1);
    }

  DisasmP(Buffer, IpcBase, Ipc, JTab, Sp);

  fprintf(Out,"s%d p%d o%d	%s	Stack: %s\n",
	  MemRd(Seg)&0xff, MemRd(JTab)&0xff, Ipc, Buffer, StackBuf);

  fflush(Out);
}

void CheckCallStack(void)
{
  int i;
  word	p=Mp;

  for (i=0;i<Level;i++)
    {
      assert(p);
      p=MemRd(WordIndexed(p, MS_DYN));
    }
  assert(p==(0xb000-4));
}

void Processor(void)
{
  byte	Opcode;
  word	w;
  float	f;
  register word	p1, p2;

  setjmp(ProcessNextInstrunction);
  for ( ; /* ever */ ; )
    {
      /*      CheckCallStack();*/
      if (Out)
	Tracer();

      CurrentIpc = Ipc;
      Opcode = FetchUB();		/* fetch next instruction */
      /* One-word load and stores constant */
      if (Opcode<128)
	{
	  p1 = Opcode;
	  TRACE(("SLDC	%d", p1));
	  Push( Opcode );		/* SLDC 0..127 Short LoaD Constant */
	}
      else
	switch (Opcode)
	  {
	  case 159:			/* LDCN LoaD Constant Nil */
	    TRACE(("LDCN	"));
	    Push( NIL );
	    break;
	  case 199:			/* LDCI LoaD Constant Integer */
	    p1=FetchW();
	    TRACE(("LDCI	%d",p1));
	    if (p1 == 16607)		/* Apple-Hack */
	      if (AppleHack1())
		Push(4);
	      else
		Push( p1 );
	    else if (p1 == 16606)
	      if (AppleHack2())
#ifdef WORD_MEMORY
		Push(Boolean(0));
#else
		Push(Boolean(0));
#endif
	      else
		Push( p1 );
	    else
	      Push( p1 );
	    break;

			/* One-word load and stores local */
					/* SLDL Short LoaD Local 1..16 */
	  case 216:	case 217:	case 218:	case 219:
	  case 220:	case 221:	case 222:	case 223:
	  case 224:	case 225:	case 226:	case 227:
	  case 228:	case 229:	case 230:	case 231:
	    p1=Opcode-216+1;
	    TRACE(("SLDL	%d", p1));
	    Push( MemRd( LocalAddr( p1 )));
	    break;
	  case 202:			/* LDL LoaD Local */
	    p1=FetchB();
	    TRACE(("LDL	%d", p1));
	    Push( MemRd( LocalAddr( p1 ) ) );
	    break;
	  case 198:			/* LLA Load Local Addres */
	    p1=FetchB();
	    TRACE(("LLA	%d", p1));
	    Push( LocalAddr( p1 ) );
	    break;
	  case 204:			/* STL STore Local */
	    p1=FetchB();
	    TRACE(("STL	%d", p1));
	    MemWr( LocalAddr( p1 ), Pop() );
	    break;

			/* One-word load and stores global */
					/* SLDO Short LoaD glObal word */
	  case 232:	case 233:	case 234:	case 235:
	  case 236:	case 237:	case 238:	case 239:
	  case 240:	case 241:	case 242:	case 243:
	  case 244:	case 245:	case 246:	case 247:
	    p1=Opcode-232+1;
	    TRACE(("SLDO	%d", p1));
	    Push( MemRd( GlobalAddr( p1 )));
	    break;
	  case 169:			/* LDO LoaD glObal */
	    p1=FetchB();
	    TRACE(("LDO	%d", p1));
	    Push( MemRd( GlobalAddr( p1 )));
	    break;
	  case 165:			/* LOA Load glObal Addres */
	    p1=FetchB();
	    TRACE(("LAO	%d", p1));
	    Push( GlobalAddr( p1 ));
	    break;
	  case 171:			/* SRO StoRe glObal */
	    p1=FetchB();
	    TRACE(("SRO	%d",p1));
	    MemWr( GlobalAddr( p1 ), Pop() );
	    break;

	    /* One-word load and stores intermediate */
	  case 182:			/* LOD LOaD */
	    p1 = FetchUB(); p2 = FetchB();
	    TRACE(("LOD	%d,%d", p1, p2));
	    Push( MemRd( IntermediateAddr( p2, p1 )));
	    break;
	  case 178:			/* LDA LOad Addres */
	    p1 = FetchUB(); p2 = FetchB();
	    TRACE(("LDA	%d,%d", p1, p2));
	    Push( IntermediateAddr( p2, p1 ));
	    break;
	  case 184:			/* STR StoRe */
	    p1 = FetchUB(); p2 = FetchB();
	    TRACE(("STR	%d,%d",p1, p2));
	    MemWr( IntermediateAddr( p2, p1 ), Pop());
	    break;

			/* One-word load and stores indirect */
					/* SIND Short INDirect */
	  case 248:	case 249:	case 250:	case 251:
	  case 252:	case 253:	case 254:	case 255:
	    p1=Opcode-248;
	    TRACE(("SIND	%d",p1));
	    Push( MemRd( WordIndexed( Pop(), p1 )));
	    break;
	  case 163:			/* IND INDirect */
	    p1=FetchB();
	    TRACE(("IND	%d",p1));
	    Push( MemRd( WordIndexed( Pop(), p1 )));
	    break;
	  case 154:			/* STO STOre indirect */
	    TRACE(("STO"));
	    p1 = Pop(); p2 = Pop();
	    PointerCheck(p2); MemWr( p2, p1);
	    break;
		
			/* One-word load and stores indirect */
	  case 157:			/* LDE LoaD Extended */
	    p1 = FetchUB(); p2 = FetchB();
	    TRACE(("LDE	%d,%d", p1, p2));
	    Push( MemRd( ExtendedAddr( p2, p1 )));
	    break;
	  case 167:			/* LAE Load Addres Extended */
	    p1 = FetchUB(); p2 = FetchB();
	    TRACE(("LAE	%d,%d", p1, p2    ));
	    Push( ExtendedAddr( p2, p1 ));
	    break;
	  case 209:			/* STE STore Extended */
	    p1 = FetchUB(); p2 = FetchB();
	    TRACE(("STE	%d,%d", p1, p2));
	    MemWr( ExtendedAddr( p2, p1 ), Pop() );
	    break;
				/* multiple-word loads and stores */
	  case 179:
	    p1=FetchUB();
	    Ipc=(Ipc+1)&(~1);		/* Nur auf Wortgrenze erlaubt */
#ifdef WORD_MEMORY
	    w=IpcBase+Ipc/2;
#else
	    w=IpcBase+Ipc;
#endif
	    TRACE(("LDC	%d%s",p1, MultipleWords(w, p1)));
	    while (p1--)
	      Push(FetchW());
	    break;
	  case 188:
	    p1=FetchUB();
	    TRACE(("LDM	%d",p1));
	    w=Pop();
	    while (p1--)
	      Push(MemRd(WordIndexed(w, p1)));
	    break;
	  case 189:
	    p1=FetchUB();
	    TRACE(("STM	%d",p1));
	    w=MemRd(WordIndexed(Sp,p1));
	    while (p1--)
	      {
		MemWr(w,Pop());
		w=WordIndexed(w,1);
	      }
	    Pop();
	    break;
				/* byte array handling */
	  case 190:
	    TRACE(("LDB	"));
	    w=Pop();
	    Push(MemRdByte(Pop(), w));
	    break;
	  case 191:
	    TRACE(("STB	"));
	    p1=Pop();
	    w=Pop();
	    MemWrByte(Pop(), w, p1);
	    break;
				/* string handling */
	  case 166:
	    assert(!(Ipc&1));
#ifdef WORD_MEMORY
	    w=IpcBase+Ipc/2;
#else
	    w=IpcBase+Ipc;
#endif
	    p1=FetchUB();
	    TRACE(("LSA	%d,\"%s\"", p1, PStr(w)));
	    Push(w);
	    Ipc+=p1;
	    break;
	  case 170:
	    p1=FetchUB();
	    TRACE(("SAS	%d",p1));
	    if ((w=Pop())&0xff00)
	      {			/* copy String */
		byte Len=MemRdByte(w, 0);
		word Dest=Pop();
		if (Len>p1)
		  XeqError(13);
		MoveLeft(Dest, 0, w, 0, Len+1);
	      }
	    else
	      {			/* store Char */
		word Dest=Pop();
		MemWrByte(Dest, 0, 1);	/* make string of len 1		    */
		MemWrByte(Dest, 1, w);	/* containing char on stack	    */
	      }
	    break;
	  case 155:
	    TRACE(("IXS	"));
	    p1=Pop();
	    p2=Pop();
	    Push(p2);
	    Push(p1);
	    if (p1>MemRdByte(p2, 0))
	      XeqError(1);
	    break;
				/* record and array handling */
	  case 168:
	    p1=FetchB();
	    TRACE(("MOV	%d",p1));
	    {
	      word	Src=Pop();
	      word	Dst=Pop();
	      while (p1--)
		{
		  MemWr(Dst, MemRd(Src));
		  Dst=WordIndexed(Dst,1);
		  Src=WordIndexed(Src,1);
		}
	    }
	    break;
	  case 162:
	    p1=FetchB();
	    TRACE(("INC	%d",p1));
	    Push(WordIndexed(Pop(),p1));
	    break;
	  case 164:
	    p1=FetchB();
	    TRACE(("IXA	%d",p1));
	    w=Pop();
	    Push(WordIndexed(Pop(),w*p1));
	    break;
	  case 192:
	    p1 = FetchUB(); p2 = FetchUB();
	    TRACE(("IXP	%d,%d", p1, p2));
	    w=Pop();
	    Push(WordIndexed(Pop(),w/p1)); /* Address */
	    Push(p2);
#ifdef IXP_COMPATIBILITY
	    Push((w%p1)*p2*0x101);
#else
	    Push((w%p1)*p2);
#endif
	    break;
	  case 208:
	    p1=FetchB();
	    TRACE(("LPA	%d,...",p1));
	    Push(IpcBase+Ipc);
	    Ipc+=p1;
	    break;
	  case 186:
	    TRACE(("LDP	"));
	    {
	      word Offset=Pop()&0xff;
	      word Size=Pop();
	      word Addr=Pop();
	      if (Offset+Size>16)
		{
		  warning("LDP: Offset(%d)+Size(%d) > Bits per word",
			  Offset, Size);
		  XeqError(1);
		}
	      Push((MemRd(Addr)>>Offset)&((1<<Size)-1));
	    }
	    break;
	  case 187:
	    TRACE(("STP	"));
	    w=Pop();
	    {
	      word Offset=Pop()&0xff;
	      word Size=Pop();
	      word Addr=Pop();
	      if (Offset+Size>16)
		{
		  warning("STP: Offset(%d)+Size(%d) > Bits per word",
			  Offset, Size);
		  XeqError(1);
		}
	      if (w>=(1<<Size))
		panic("STP: Value %d cannot be stored in %d bits", w, Size);
	      MemWr(Addr, 
		    (MemRd(Addr) & ~(((1<<Size)-1)<<Offset)) | (w<<Offset));
	    }
	    break;
				/* TOS arithmetic: integers */
	  case 128:			/* ABI ABsolute Integer */
	    TRACE(("ABI	"));
	    Push( abs( PopInteger()));
	    break;
	  case 130:
	    TRACE(("ADI	"));
	    Push( PopInteger() + PopInteger() );
	    break;
	  case 145:
	    TRACE(("NGI	"));
	    Push( -PopInteger() );
	    break;
	  case 149:
	    TRACE(("SBI	"));
	    {
	      Integer i=PopInteger();
	      Push( PopInteger()-i );
	    }
	    break;
	  case 143:
	    TRACE(("MPI	"));
	    Push( PopInteger() * PopInteger() );
	    break;
	  case 152:
	    TRACE(("SQI	"));
	    {
	      Integer i=PopInteger();
	      Push( i*i );
	    }
	    break;
	  case 134:
	    TRACE(("DVI	"));
	    {
	      Integer i=PopInteger();
	      if (!i)
		XeqError(6);
	      Push( PopInteger() / i );
	    }
	    break;
	  case 142:
	    TRACE(("MODI	"));
	    {
	      Integer i=PopInteger(); 
	      if (!i)
		XeqError(6);
	      Push( PopInteger() % i );
	    }
	    break;
	  case 136:
	    TRACE(("CHK	"));
	    {
	      Integer Upper=PopInteger();
	      Integer Lower=PopInteger();
	      Integer Value=PopInteger();
	      Push(Value);
	      if ( (Value>Upper) || (Value<Lower) )
		XeqError(1);
	    }
	    break;
	  case 195:
	    TRACE(("EQUI	"));
	    {
	      Integer i=PopInteger();
	      Push ( Boolean ( PopInteger() == i ) );
	    }
	    break;
	  case 203:
	    TRACE(("NEQI	"));
	    {
	      Integer i=PopInteger();
	      Push ( Boolean ( PopInteger() != i ) );
	    }
	    break;
	  case 200:
	    TRACE(("LEQI	"));
	    {
	      Integer i=PopInteger();
	      Push ( Boolean ( PopInteger() <= i ) );
	    }
	    break;
	  case 201:
	    TRACE(("LESI	"));
	    {
	      Integer i=PopInteger();
	      Push ( Boolean ( PopInteger() < i ) );
	    }
	    break;
	  case 196:
	    TRACE(("GEQI	"));
	    {
	      Integer i=PopInteger();
	      Push ( Boolean ( PopInteger() >= i ) );
	    }
	    break;
	  case 197:
	    TRACE(("GRTI	"));
	    {
	      Integer i=PopInteger();
	      Push ( Boolean ( PopInteger() > i ) );
	    }
	    break;
				/* TOS arithmetic: reals */
	  case 138:
	    TRACE(("FLT	"));
	    PushReal(PopInteger());
	    break;
	  case 137:
	    TRACE(("FLO	"));
	    f=PopReal();
	    PushReal(PopInteger());
	    PushReal(f);
	    break;
	  case 129:
	    TRACE(("ABR	"));
	    PushReal(fabs(PopReal()));
	    break;
	  case 131:
	    TRACE(("ADR	"));
	    PushReal(PopReal()+PopReal());
	    break;
	  case 146:
	    TRACE(("NGR	"));
	    PushReal(-PopReal());
	    break;
	  case 150:
	    TRACE(("SBR	"));
	    f=PopReal();
	    PushReal(PopReal()-f);
	    break;
	  case 144:
	    TRACE(("MPR	"));
	    PushReal(PopReal()*PopReal());
	    break;
	  case 153:
	    TRACE(("SQR	"));
	    f=PopReal();
	    PushReal(f*f);
	    break;
	  case 135:
	    TRACE(("DVR	"));
	    f=PopReal();
	    PushReal(PopReal()/f);
	    break;

	  case 175:
	    p1=FetchUB();
	    switch(p1)
	      {
	      case 2:
		TRACE(("EQUREAL	"));
		f=PopReal();
		Push(Boolean( PopReal() == f ));
		break;
	      case 4:
		TRACE(("EQUSTR	"));
		w=Pop();
		Push(Boolean(StrCmp(Pop(), w) == 0));
		break;
	      case 6:
		TRACE(("EQUBOOL	"));
		Push(Boolean((Pop()&1)==(Pop()&1)));
		break;
	      case 8:
		TRACE(("EQUPOWR	"));
		{
		  Set_t Set1;
		  Set_t Set2;
		  SetPop(&Set1);
		  SetPop(&Set2);
		  Push(Boolean(SetCmp(&Set1, &Set2) == 0));
		  break;
		}
	      case 10:
		p1=FetchUB();
		TRACE(("EQUBYT	%d", p1));
		w=Pop();
		Push(Boolean(ByteCmp(Pop(), w, p1) == 0));
		break;
	      case 12:
		p1=FetchUB();
		TRACE(("EQUWORD	%d", p1));
		assert(!(p1&1));
		w=Pop();
		Push(Boolean(WordCmp(Pop(), w, p1/2) == 0));
		break;
	      default:
		TRACE(("EQU	%d",p1));
		panic("EQU: Unsupported data type %d", p1);
	      }
	    break;
	  case 183:
	    p1=FetchUB();
	    switch(p1)
	      {
	      case 2:
		TRACE(("NEQREAL	"));
		f=PopReal();
		Push(Boolean( PopReal() != f ));
		break;
	      case 4:
		TRACE(("NEQSTR	"));
		w=Pop();
		Push(Boolean(StrCmp(Pop(), w) != 0));
		break;
	      case 6:
		TRACE(("NEQBOOL	"));
		Push(Boolean((Pop()&1)!=(Pop()&1)));
		break;
	      case 8:
		TRACE(("NEQPOWR	"));
		{
		  Set_t Set1;
		  Set_t Set2;
		  SetPop(&Set1);
		  SetPop(&Set2);
		  Push(Boolean(SetCmp(&Set1, &Set2) != 0));
		  break;
		}
	      case 10:
		p1=FetchUB();
		TRACE(("NEQBYT	%d", p1));
		w=Pop();
		Push(Boolean(ByteCmp(Pop(), w, p1) != 0));
		break;
	      case 12:
		p1=FetchUB();
		TRACE(("NEQWORD	%d", p1));
		assert(!(p1&1));
		w=Pop();
		Push(Boolean(WordCmp(Pop(), w, p1/2) != 0));
		break;
	      default:
		TRACE(("NEQ	%d",p1));
		panic("NEQ: Unsupported data type %d", p1);
	      }
	    break;

	  case 180:
	    p1=FetchUB();
	    switch(p1)
	      {
	      case 2:
		TRACE(("LEQREAL	"));
		f=PopReal();
		Push(Boolean( PopReal() <= f ));
		break;
	      case 4:
		TRACE(("LEQSTR	"));
		w=Pop();
		Push(Boolean(StrCmp(Pop(), w) <= 0) );
		break;
	      case 6:
		TRACE(("LEQBOOL	"));
		w=Pop()&1;
		Push(Boolean((Pop()&1) <= w));
		break;
	      case 8:
		TRACE(("LEQPOWR	"));
		{
		  Set_t Set1;
		  Set_t Set2;
		  SetPop(&Set1);
		  SetPop(&Set2);
		  Push(Boolean(SetIsSubset(&Set1, &Set2)));
		  break;
		}
		break;
	      case 10:
		p1=FetchUB();
		TRACE(("LEQBYT	%d", p1));
		w=Pop();
		Push(Boolean(ByteCmp(Pop(), w, p1) <= 0));
		break;
	      default:
		TRACE(("LEQ	%d",p1));
		panic("LEQ: Unsupported data type %d", p1);
	      }
	    break;

	  case 181:
	    p1=FetchUB();
	    switch(p1)
	      {
	      case 2:
		TRACE(("LESREAL	"));
		f=PopReal();
		Push(Boolean( PopReal() < f ));
		break;
	      case 4:
		TRACE(("LESSTR	"));
		w=Pop();
		Push(Boolean(StrCmp(Pop(), w) < 0) );
		break;
	      case 6:
		TRACE(("LESBOOL	"));
		w=Pop()&1;
		Push(Boolean((Pop()&1) < w));
		break;
	      case 10:
		p1=FetchUB();
		TRACE(("LESBYT	%d", p1));
		w=Pop();
		Push(Boolean(ByteCmp(Pop(), w, p1) < 0));
		break;
	      default:
		TRACE(("LES	%d",p1));
		panic("LES: Unsupported data type %d", p1);
	      }
	    break;

	  case 176:
	    p1=FetchUB();
	    switch(p1)
	      {
	      case 2:
		TRACE(("GEQREAL	"));
		f=PopReal();
		Push(Boolean( PopReal() >= f ));
		break;
	      case 4:
		TRACE(("GEQSTR	"));
		w=Pop();
		Push(Boolean(StrCmp(Pop(), w) >= 0) );
		break;
	      case 6:
		TRACE(("GEQBOOL	"));
		w=Pop()&1;
		Push(Boolean((Pop()&1) >= w));
		break;
	      case 8:
		TRACE(("GEQPOWR	"));
		{
		  Set_t Set1;
		  Set_t Set2;
		  SetPop(&Set1);
		  SetPop(&Set2);
		  Push(Boolean(SetIsSubset(&Set2, &Set1)));
		  break;
		}
		break;
	      case 10:
		p1=FetchUB();
		TRACE(("GEQBYT	%d", p1));
		w=Pop();
		Push(Boolean(ByteCmp(Pop(), w, p1) >= 0));
		break;
	      default:
		TRACE(("GEQ	%d",p1));
		panic("GEQ: Unsupported data type %d", p1);
	      }
	    break;

	  case 177:
	    p1=FetchUB();
	    switch(p1)
	      {
	      case 2:
		TRACE(("GRTREAL	"));
		f=PopReal();
		Push(Boolean( PopReal() > f ));
		break;
	      case 4:
		TRACE(("GRTSTR	"));
		w=Pop();
		Push(Boolean(StrCmp(Pop(), w) > 0) );
		break;
	      case 6:
		TRACE(("GRTBOOL	"));
		w=Pop()&1;
		Push(Boolean((Pop()&1) > w));
		break;
	      case 10:
		p1=FetchUB();
		TRACE(("GRTBYT	%d", p1));
		w=Pop();
		Push(Boolean(ByteCmp(Pop(), w, p1) > 0));
		break;
	      default:
		TRACE(("GRT	%d",p1));
		panic("GRT: Unsupported data type %d", p1);
	      }
	    break;
				/* TOS arithmetic: logical */
	  case 132:
	    TRACE(("LAND	"));
	    Push( Pop() & Pop() );
	    break;
	  case 141:
	    TRACE(("LOR	"));
	    Push( Pop() | Pop() );
	    break;
	  case 147:
	    TRACE(("LNOT	"));
	    Push( ~Pop() );
	    break;
				/* Sets */
	  case 160:
	    p1=FetchUB();
	    TRACE(("ADJ	%d",p1));
	    w=MemRd(Sp);
	    if (p1!=w)
	      {
		Set_t Buf;
		SetPop(&Buf);
		SetAdj(&Buf, p1);
		SetPush(&Buf);
	      }
	    if (p1!=Pop())
	      panic("adj failure");
	    break;
	  case 151:
	    TRACE(("SGS	"));
	    w=Pop();
	    if (w<512)
	      {
		int	Size=(w+16)/16;
		word	Addr;
		int	i;
		for (i=0;i<Size;i++)
		  Push(0);
		Addr=WordIndexed(Sp, w/16);
		MemWr(Addr, MemRd(Addr)|(1<<w%16));
		Push(Size);
	      }
	    else
	      XeqError(1);
	    break;
	  case 148:
	    TRACE(("SRS	"));
	    p1=Pop();
	    p2=Pop();
	    if ((p1<512) && (p1<512) )
	      {
		if (p2>p1)
		  Push(0);
		else
		  {
		    int	Size=(p1+16)/16;
		    word	Addr;
		    int	i;
		    for (i=0;i<Size;i++)
		      Push(0);
		    while (p2<=p1)
		      {
			Addr=WordIndexed(Sp, p2/16);
			MemWr(Addr, MemRd(Addr)|(1<<p2%16));
			p2++;
		      }
		    Push(Size);
		  }
	      }
	    else
	      XeqError(1);	    
	    break;
	  case 139:
	    TRACE(("INN	"));
	    {
	      word Size;
	      word Addr;
	      word Val;
	      Size=Pop();
	      Addr=Sp;
	      Sp=WordIndexed(Sp,Size);
	      Val=Pop();
	      if (Val>=16*Size)
		Push(Boolean(0));
	      else
		Push(Boolean(MemRd(WordIndexed(Addr,(Val/16))) &
			     (1<<(Val%16))));
	    }
	    break;
	  case 156:
	    TRACE(("UNI	"));
	    {
	      int   i;
	      word  Size;
	      Set_t Set;

	      SetPop(&Set);
	      Size=Pop();
	      if (Size>Set.Size)
		SetAdj(&Set, Size);

	      for (i=0; i<Size; i++)
		Set.Data[i]|=Pop();
	      SetPush(&Set);
	    }
	    break;
	  case 140:
	    TRACE(("INT	"));
	    {
	      int   i;
	      word  Size;
	      Set_t Set;

	      SetPop(&Set);
	      Size=Pop();
	      if (Size>Set.Size)
		SetAdj(&Set, Size);

	      for (i=0; i<Size; i++)
		Set.Data[i]&=Pop();
	      while(i<Set.Size)
		Set.Data[i++]=0;
	      SetPush(&Set);
	    }
	    break;

	  case 133:
	    TRACE(("DIF	"));
	    {
	      int   i;
	      word  Size;
	      Set_t Set;

	      SetPop(&Set);
	      Size=Pop();
	      if (Size>Set.Size)
		SetAdj(&Set, Size);

	      for (i=0; i<Size; i++)
		Set.Data[i] = Pop() & ~Set.Data[i];
	      while (i<Set.Size)
		Set.Data[i]=0;
	      SetPush(&Set);
	    }
	    break;
				/* jumps */
	  case 185:
	    p1=FetchUB();
	    TRACE(("UJP	%d",(signed char)p1));
	    w=jump((signed char)p1);
	    if ( ( Ipc-w == 5 ) &&	/* check for endless loop */
		 ( MemRdByte(IpcBase, w  ) == 1   ) && /* SLDC 1 */
		 ( MemRdByte(IpcBase, w+1) == 161 ) && /* FJP  2 */
		 ( MemRdByte(IpcBase, w+2) == 2   ) )
	      sleep(1);			/* reduce processor load */
	    Ipc=w;
	    break;
	  case 161:
	    p1=FetchUB();
	    TRACE(("FJP	%d",(signed char)p1));
	    if (! (Pop()&1) )
	      {
		w=jump((signed char)p1);
		if ( ( Ipc-w == 3 ) &&	/* check for endless loop */
		     ( MemRdByte(IpcBase, w  ) == 0 ) ) /* SLDC 0 */
		  sleep(1);			/* reduce processor load */
		Ipc=w;
	      }
	    break;
	  case 211:
	    p1=FetchUB();
	    TRACE(("EFJ	%d",(signed char)p1));
	    if ( Pop() != Pop() )
	      Ipc=jump((signed char)p1);
	    break;
	  case 212:
	    p1=FetchUB();
	    TRACE(("NFJ	%d",(signed char)p1));
	    if ( Pop() == Pop() )
	      Ipc=jump((signed char)p1);
	    break;
	  case 172:
	    Ipc=(Ipc+1)&(~1);
	    p1=FetchW(); p2=FetchW();
	    TRACE(("XJP	%d,%d",p1,p2));
	    w=Pop();
	    if ( (w>=p1) && (w<=p2) )
	      {
		Ipc=Ipc+2*(w-p1)+2;
#ifdef WORD_MEMORY
		Ipc-=MemRd(IpcBase+Ipc/2);
#else
		Ipc-=MemRd(IpcBase+Ipc);
#endif
	      }
	    break;
				/* procedure and function calls */
	  case 206:
	    p1=FetchUB();
	    TRACE(("CLP	%d",p1));
	    call(Seg, p1, Mp);
	    break;
	  case 207:
	    p1=FetchUB();
	    TRACE(("CGP	%d",p1));
	    call(Seg, p1, Base);
	    break;
	  case 174:
	    p1=FetchUB();
	    TRACE(("CIP	%d",p1));
	    call(Seg, p1, StaticLink(Seg, p1));
	    break;
	  case 194:
	    p1=FetchUB();
	    TRACE(("CBP	%d",p1));
	    call(Seg, p1, InitialMp);
	    break;
	  case 205:
	    p1=FetchUB(); p2=FetchUB();
	    TRACE(("CXP	%d,%d", p1, p2));
	    if (p1)			/* Nicht bei Segment 0		    */
	      CspLoadSegment(p1);
	    w=SegDict[p1].Seg;
	    if (call(w, p2, StaticLink(w, p2) ))
	      CspUnloadSegment(p1);
	    break;
	  case 173:
	    p1=FetchUB();
	    TRACE(("RNP	%d",p1));
	    ret(p1);
	    break;
	  case 193:
	    p1=FetchUB();
	    TRACE(("RBP	%d",p1));
	    Base=Pop();
	    MemWr(STKBASE, Base);
	    if ( (Base<Kp) || (Base>InitialMp))
	      warning("RBP: Base %04x out of range", Base);

	    ret(p1);
	    break;

	  case 158:		/* CSP, Call Standard Procedure */
	    p1=FetchUB();
	    switch(p1)
	      {
	      case 0:
		TRACE(("CSP	IoCheck"));
		if (MemRd(IORSLT))
		  XeqError(10);
		break;
	      case 1:
		TRACE(("CSP	new"));
		ClrGDirP();
		w=Pop();
		MemWr(Pop(),Np);
		Np=WordIndexed(Np, w);
		StackCheck();
		break;
	      case 2:
		TRACE(("CSP	Moveleft"));
		{
		  word    Len       = Pop();
		  Integer DstOffset = PopInteger();
		  word    Dst       = Pop();
		  Integer SrcOffset = PopInteger();
		  word    Src       = Pop();
		  MoveLeft(Dst, DstOffset, Src, SrcOffset, Len);
		}
		break;
	      case 3:
		TRACE(("CSP	Moveright"));
		{
		  word    Len       = Pop();
		  Integer DstOffset = PopInteger();
		  word    Dst       = Pop();
		  Integer SrcOffset = PopInteger();
		  word    Src       = Pop();
		  MoveRight(Dst, DstOffset, Src, SrcOffset, Len);
		}
		break;
	      case 4:
		TRACE(("CSP	exit"));
		{
		  word	ProcNo=Pop();
		  word	SegNo=Pop();
		  word	xMp=Mp;
		  word	xSeg=Seg;
		  word	xJTab=JTab;

		  Ipc=ProcExitIpc(xJTab);
		  while ( (ProcNumber(xJTab) != ProcNo) ||
			  (SegNumber(xSeg)   != SegNo) )
		    {
		      if (!xMp ||
			  !(xJTab = MemRd(WordIndexed(xMp, MS_JTAB))) ||
			  !(xSeg  = MemRd(WordIndexed(xMp, MS_SEG))) )
			XeqError(3);

		      MemWr(WordIndexed(xMp, MS_IPC), ProcExitIpc(xJTab));
		      xMp   = MemRd(WordIndexed(xMp, MS_DYN));
		    }
		}
		break;
	      case 5:
		TRACE(("CSP	unitread"));
		{
		  word p1, p2, p3, p4, p5, p6;
		  p6=Pop();p5=Pop();p4=Pop();
		  p3=Pop();p2=Pop();p1=Pop();
		  UnitRead(p1,p2,p3,p4,p5,p6);
		}
		break;
	      case 6:
		TRACE(("CSP	unitwrite"));
		{
		  word p1, p2, p3, p4, p5, p6;
		  p6=Pop();p5=Pop();p4=Pop();
		  p3=Pop();p2=Pop();p1=Pop();
#ifdef WORD_MEMORY
		  if ( (SegNumber(Seg)==0) &&
		       (ProcNumber(JTab)==20) )
		    p3--;		/* APPLE-Hack */
#endif
		  UnitWrite(p1,p2,p3,p4,p5,p6);
		}
		break;
	      case 9:
		TRACE(("CSP	time"));
		MemWr(Pop(),0);
		MemWr(Pop(),0);
		break;
	      case 10:
		TRACE(("CSP	fillchar"));
		{
		  word    ch     = Pop();
		  word    Len    = Pop();
		  Integer Offset = PopInteger();
		  word    Addr   = Pop();
		  if (!(Len&0x8000))
		    while (Len--)
		      MemWrByte(Addr, Offset++, ch);
		}
		break;
	      case 11:
		TRACE(("CSP	scan"));
		{
		  word	  Dummy  = Pop();
		  Integer Offset = PopInteger();
		  word    Buf    = Pop();	/* Buffer Address */
		  word    ch     = Pop();	/* zu suchendes Zeichen */
		  word    match  = Pop();	/* 0 suche nach ==ch,
						   !=0: Suche nach !=ch */
		  word    limit  = Pop();	/* Limit */
		  word	  res;

		  if (limit&0x8000)
		    {
		      limit=0x10000-limit;
		      for (res=0; res<limit; res++)
			if (MemRdByte(Buf, Offset-res)!=ch)
			  {
			    if (match)
			    break;
			  }
			else
			  {
			    if (!match)
			      break;
			  }
		      Push(0x10000-res);
		    }
		  else
		    {
		      for (res=0; res<limit; res++)
			if (MemRdByte(Buf, Offset+res)!=ch)
			  {
			    if (match)
			      break;
			  }
			else
			  {
			    if (!match)
			      break;
			  }
		      Push(res);
		    }
		}
		break;
	      case 12:
		TRACE(("CSP	unitstat"));
		{
		  word    Dummy  = Pop();
		  Integer Offset = PopInteger();
		  word    Addr   = Pop();
		  word    Unit   = Pop();
		  UnitStat(Unit, Addr, Offset, Dummy);
		}
		break;
	      case 21:
		TRACE(("CSP	load_segment"));
		CspLoadSegment(Pop());
		break;
	      case 22:
		TRACE(("CSP	unload_segment"));
		CspUnloadSegment(Pop());
		break;
	      case 23:
		TRACE(("CSP	TruncReal"));
		f=PopReal();
		if (f<0)
		  Push(ceil(f));
		else
		  Push(floor(f));
		break;
	      case 24:
		TRACE(("CSP	RoundReal"));
		Push(rint(PopReal()));
		break;
	      case 32:
		TRACE(("CSP	mark"));
		ClrGDirP();
		MemWr(Pop(),Np);
		break;
	      case 33:
		TRACE(("CSP	release"));
		Np=MemRd(Pop());
		StackCheck();
		MemWr(GDIRP, NIL);
		break;
	      case 34:
		TRACE(("CSP	ioresult"));
		Push(MemRd(IORSLT));
		break;
	      case 35:
		TRACE(("CSP	unitbusy"));
		Push(UnitBusy(Pop()));
		break;
	      case 36:
		TRACE(("CSP	PwrOfTen"));
		{
		  float PwrOfTen[]={1e0,  1e1,  1e2,  1e3,  1e4,  1e5,
				    1e6,  1e7,  1e8,  1e9,  1e10, 1e11,
				    1e12, 1e13, 1e14, 1e15, 1e16, 1e17,
				    1e18, 1e19, 1e20, 1e21, 1e22, 1e23,
				    1e24, 1e25, 1e26, 1e27, 1e28, 1e29,
				    1e30, 1e31, 1e32, 1e33, 1e34, 1e35,
				    1e36, 1e37, 1e38, 1e39};
		  int Value=PopInteger();
		  if ( (Value<0) || (Value>39) )
		    PushReal(0);
		  else
		    PushReal(PwrOfTen[Value]);
		}
		break;
	      case 37:
		TRACE(("CSP	unitwait"));
		UnitWait(Pop());
		break;
	      case 38:
		TRACE(("CSP	unitclear"));
		UnitClear(Pop());
		break;
	      case 39:
		TRACE(("CSP	halt"));
		return;
		break;
	      case 40:
		TRACE(("CSP	memavail"));
		if (MemRd(GDIRP))
		  w=Kp-MemRd(GDIRP);
		else
		  w=Kp-Np;
		Push(w/2);
		break;
	      default:
		p2=FetchUB();
		TRACE(("CSP	%d (unknown)", p1));
		warning("calling unimplemented standard procedure %d", p1);
		XeqError(11);
	      }
	    break;

	  case 213:
	    TRACE(("BPT	"));
	    XeqError(14);
	    break;
	  case 214:
	    TRACE(("XIT	"));
	    return;
	    break;
	  case 215:
	    TRACE(("NOP	"));
	    break;
	  default:
	    p1=FetchUB();p2=FetchUB();
	    TRACE(("db	%d, %d, %d, %d ...",Opcode, p1, p2, FetchUB()));
	    warning("unknown instuction %d at s%d p%d o%d",
		    Opcode, SegNumber(Seg), ProcNumber(JTab), CurrentIpc);
	    XeqError(11);
	    break;
	  }
    }
}

word LookupFile(word Unit, const char *Name)
{
  int i;
  UnitRead(Unit, Np, 0, 2048, 2, 0);
  if (MemRd(IORSLT))
    return(0);

  for (i=0;i<MemRd(WordIndexed(Np, 8));i++)
    {
      word Entry=WordIndexed(Np, 13+13*i);
      int len;
      for (len=0; len<MemRdByte(WordIndexed(Entry, 3),0); len++)
	if (toupper(MemRdByte(WordIndexed(Entry, 3),1+len)) !=
	    toupper(Name[len]))
	  goto next;
      if (Name[len])
	continue;
      return(MemRd(WordIndexed(Entry,0)));
    next:
    }
  return(0);
}

/* Das Segment 0 ist aufgespalten, und die Zeiger im
   Procedure-Dictionary sind so korrigiert worden, dass nach dem Laden
   der beiden Hlften an die jeweils "richtige" Addresse die Zeiger
   korrekt sind.

   Diese Routine korrigiert die Zeiger im Segment-Dictionary. Dazu
   ermittelt sie zuerst die Addresse, an der die zwiete hlfte
   eigentlich geladen werden sollte. Danach wird ein Offset ermittelt,
   mit dem die Zeiger in die zweite Hlfte korrigiert werden mssen. */

static void FixupSeg0(word LoadAddr)
{
  word Seg=SegDict[0].Seg;
  word SegBase=SegDict[0].SegBase;
  word Addr;
  word Offset;
  int i;

  Addr=0;
  for (i=1;i<=SegNumProc(Seg); i++)
    {
      word JTab=Proc(Seg, i);
      if (JTab<SegBase)
	if (JTab>Addr)
	  Addr=JTab;
    }
  if (!Addr)
    return;			/* no Fixup needed */
  Addr=WordIndexed(Addr,1);
  Offset=LoadAddr-Addr;
  if (!Offset)
    return;

  for (i=1;i<=SegNumProc(Seg); i++)
    {
      word JTab=Proc(Seg, i);
      if ( (JTab<SegBase) )
	{
	  Addr=WordIndexed(Seg,-i);
#ifdef WORD_MEMORY
	  MemWr(Addr, MemRd(Addr)-2*Offset);
#else
	  MemWr(Addr, MemRd(Addr)-Offset);
#endif
	}
    }
}

static void load(word Unit, word BlockNo)
{
  int	i;

  UnitRead(Unit, Np, 0, 512, BlockNo, 0);
  if (MemRd(IORSLT))
    return;

  /* Erzeuge das Segment Dictionary */
  for (i=0;i<16;i++)
    {
      word	CodeAddr=MemRd(WordIndexed(Np, 2*i))+BlockNo;
      word	CodeLeng=MemRd(WordIndexed(Np, 2*i+1));
      word	SegInfo =MemRd(WordIndexed(Np, i+0x80));

      assert (!(CodeLeng&1));

      if (CodeAddr && CodeLeng)
	{
	  int	SegNo=SegInfo&0xff;
	  if (SegInfo&0x0f00)
	    {
	      MemWr(SEG_UNIT(SegNo),  Unit);
	      MemWr(SEG_BLOCK(SegNo), CodeAddr);
	      MemWr(SEG_SIZE(SegNo),  CodeLeng);
	    }
	  if (SegNo==0)
	    {
	      if (SegInfo==0xc200)
		{
		  SegDict[0].UseCount++;
		  SegDict[0].OldKp=Kp;
		  SegDict[0].Seg=WordIndexed(Kp,-1);
#ifdef WORD_MEMORY
		  Kp-=CodeLeng/2;
#else
		  Kp-=CodeLeng;
#endif
		  SegDict[0].SegBase=Kp;
		  UnitRead(Unit, Kp, 0, CodeLeng, CodeAddr, 0);
		}
	      else if (SegInfo==0)
		{
#ifndef WORD_MEMORY
#ifdef APPLE_SEG0_LOAD_GAP
		  if (AppleCompatibility)
		    {
		      Kp-=APPLE_SEG0_LOAD_GAP;
		      assert (Syscom>=Kp );
		      assert (WordIndexed(Syscom, SYSCOM_SIZE) <
			      Kp+APPLE_SEG0_LOAD_GAP );
		    }
#endif
#endif
		  FixupSeg0(Kp);
#ifdef WORD_MEMORY
		  Kp-=CodeLeng/2;
#else
		  Kp-=CodeLeng;
#endif
		  UnitRead(Unit, Kp, 0, CodeLeng, CodeAddr, 0);
		}
	    }
	}
    }
}

int main (int argc, char *argv[])
{
  int	i;
  int	WriteMode=0;
  int	SystemUnit;
  int	RootUnit=4;
  word	SystemBlock;
#ifdef WORD_MEMORY
  char	*RootDisk="apple1.ucsd";
#else
  char	*RootDisk="apple0.ucsd";
#endif
  char	*SystemFileName="system.pascal";

  memset(SegDict, 0, sizeof(SegDict));
  MemInit();
  UnitInit();

  while ((i=getopt(argc, argv, 
#ifndef WORD_MEMORY
		   "a"
#endif
		   "gn:t:wrs23l"))!=EOF)
    switch(i)
      {
#ifndef WORD_MEMORY
      case 'a':
	AppleCompatibility=1;
	break;
#endif
      case 'g':
	TraceLevel=0;
	break;
      case 't':
	if (!optarg || !*optarg)
	  {
	    fprintf(stderr,"-t option requires filename argument\n");
	    exit(1);
	  }
	if (!(Out=fopen(optarg,"w")))
	  {
	    perror(optarg);
	    exit(1);
	  }
	break;
      case 'w':
	WriteMode=1;
	break;
      case 'r':
	RootUnit=5;
	break;
      case 's':
	RootDisk="work.ucsd";
	DiskMount(5, "system.ucsd", ReadWrite);
	break;
      case '2':
	DiskMount(9, "apple2.ucsd", ReadOnly);
	break;
      case '3':
	DiskMount(9, "apple3.ucsd", ReadOnly);
	break;
      case 'l':
	DiskMount(10, "library.ucsd", ReadWrite);
	break;
      case 'n':
	if (!optarg || !*optarg)
	  {
	    fprintf(stderr,"-n option requires filename argument\n");
	    exit(1);
	  }
	SystemFileName=optarg;
	break;
      }
  DiskMount(4, RootDisk, WriteMode?ReadWrite:Forget);
    
#ifndef WORD_MEMORY
  if (AppleCompatibility)
    {
      Np=APPLE_HEAP_BOT;
      Kp=APPLE_KP_TOP;
      Syscom=APPLE_SYSCOM;
    }
  else
#endif
    {
      Np=HEAP_BOT;
      Kp=KP_TOP;
      Kp=WordIndexed(Kp, -SYSCOM_SIZE);
      Syscom=Kp;
    }
  Sp=SP_TOP;

  SystemBlock=0;
  SystemUnit=0;
  if ((SystemBlock=LookupFile(RootUnit, SystemFileName)))
    SystemUnit=RootUnit;
  else
    for (SystemUnit=4; SystemUnit<32; SystemUnit++)
      {
	if (SystemUnit==6)
	  SystemUnit=9;
	if ((SystemBlock=LookupFile(SystemUnit, SystemFileName)))
	  break;
      }
  if (!SystemBlock || !SystemUnit)
    panic("%s: not found", SystemFileName );

  load(SystemUnit, SystemBlock);
  if (MemRd(IORSLT))
    panic("$s unit %d clock %d: Ioerror %d",
	  SystemFileName, SystemUnit, SystemBlock,MemRd(IORSLT));
  if (!SegDict[0].UseCount)
    panic("%s: not a valid system, no segment 0", SystemFileName );    

  Mp=Kp;
  call(SegDict[0].Seg, 1, NIL);
  InitialMp=Mp;
  Sp=WordIndexed(Sp,1);

  MemWr( LocalAddr(1), Syscom );
  MemWr( GDIRP   ,NIL );
  MemWr( SYSUNIT ,RootUnit );
  MemWr( WordIndexed(Syscom,161), 0x4bd);
  MemReadOnly(WordIndexed(Syscom,161), WordIndexed(Syscom,161), 1);
  MemWr( WordIndexed(Syscom,166), 0x6);
  MemWrByte( WordIndexed(Syscom,169), 1, 0x81);	/* Bitfeld mit
                                                   unbekanntem Inhalt */
  Processor();

  DumpCore();
  if (Out)
    fclose(Out);
  UnitFinish();
  return(0);
}
